@@ -78,26 +78,24 @@ static int firegl_acpi_video_event(struct notifier_block *nb, unsigned long val,
 {
     struct acpi_bus_event *entry;
     KCL_ACPI_ContextHandle ctx_handle;
-
+    unsigned int  ret = 0;
     entry = (struct acpi_bus_event *)data;
     if ( !strcmp(entry->device_class, KCL_ACPI_VIDEO_CLASS) )
     {            
-        ctx_handle = firegl_query_busid(entry->bus_id, KCL_ACPI_VIDEO_CLASS);
+        ctx_handle = firegl_query_acpi_handle((KCL_NOTIFIER_BLOCKER)nb, KCL_ACPI_VIDEO_CLASS);
         if (ctx_handle == NULL)
         {
             KCL_DEBUG_ERROR ("Could not find private acpi context by video busid: %s\n", entry->bus_id);
         }
         else
         {
-            libip_video_notify(NULL, entry->type, ctx_handle);
+            ret = libip_video_notify(NULL, entry->type, ctx_handle);
         }
     }
-    return NOTIFY_OK;
-}
 
-static struct notifier_block firegl_acpi_video_notifier = {
-        .notifier_call = firegl_acpi_video_event,
-};
+    return (ret == 0)? NOTIFY_OK:NOTIFY_BAD;
+
+}
 
 /** \brief kernel call back function when acpi ac events happen
  * \param nb Notifier block from kernel. Not used in driver.
@@ -115,7 +113,7 @@ static int firegl_acpi_ac_event(struct notifier_block *nb, unsigned long val,
     entry = (struct acpi_bus_event *)data;
     if ( !strcmp(entry->device_class, KCL_ACPI_AC_CLASS) )
     {            
-        ctx_handle = firegl_query_busid(entry->bus_id, KCL_ACPI_AC_CLASS);
+        ctx_handle = firegl_query_acpi_handle((KCL_NOTIFIER_BLOCKER)nb, KCL_ACPI_AC_CLASS);
         if (ctx_handle == NULL)
         {
             KCL_DEBUG_ERROR ("Could not find private acpi context by ac_adpater busid: %s\n", entry->bus_id);
@@ -129,19 +127,13 @@ static int firegl_acpi_ac_event(struct notifier_block *nb, unsigned long val,
     return NOTIFY_OK;
 }
 
-static struct notifier_block firegl_acpi_ac_notifier = {
-        .notifier_call = firegl_acpi_ac_event,
-};
-
-
 static int firegl_acpi_lid_event(struct notifier_block *nb, unsigned long val,
                                void *data)
 {
+    unsigned int status;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
-    if (!acpi_lid_open()) 
-    {	
-        libip_lid_open_notify();
-    }
+    status = acpi_lid_open();
+    libip_lid_open_notify(status);
 #endif
     return NOTIFY_OK;
 }
@@ -356,6 +348,23 @@ error:
     return status;
 }
 
+void* KCL_ACPI_GetVfctBios(unsigned long *size)
+{
+    struct acpi_table_header *hdr;
+    acpi_size tbl_size ;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,6,3)    
+    if (!ACPI_SUCCESS(acpi_get_table_with_size("VFCT", 1, &hdr, &tbl_size)))
+#else
+    tbl_size = 0x7fffffff;
+    if (!ACPI_SUCCESS(acpi_get_table("VFCT", 1, &hdr)))
+#endif
+    {
+        return NULL;
+    }
+    *size = tbl_size;
+    return hdr;
+}
+
 /** \brief install notify handler for acpi device
  *  \param device acpi device
  *  \param handler_type type of handler
@@ -366,7 +375,8 @@ error:
 unsigned int ATI_API_CALL KCL_ACPI_InstallHandler(KCL_ACPI_DevHandle device,
                                                   unsigned int handler_type,
                                                   KCL_ACPI_CallbackHandle handler,
-                                                  KCL_ACPI_ContextHandle context)
+                                                  KCL_ACPI_ContextHandle context,
+                                                  KCL_NOTIFIER_BLOCKER *nb)
 {
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
     return acpi_install_notify_handler(device, handler_type, (acpi_notify_handler)handler, context);
@@ -377,13 +387,27 @@ unsigned int ATI_API_CALL KCL_ACPI_InstallHandler(KCL_ACPI_DevHandle device,
     if ((KCL_ACPI_GetDevHandle(device, "_DOD", &dummy_handle) == KCL_ACPI_OK)
         || (KCL_ACPI_GetDevHandle(device, "_DOS", &dummy_handle) == KCL_ACPI_OK))
     {
-        return register_acpi_notifier(&firegl_acpi_video_notifier);
+        *nb = kmalloc(sizeof(struct notifier_block), GFP_KERNEL);
+        if (!*nb)
+        {
+            KCL_DEBUG_ERROR ("Could not allocate enough memory for video notifier_block\n");
+            return -ENOMEM;
+        }
+        ((struct notifier_block*)(*nb))->notifier_call = firegl_acpi_video_event;
+        return register_acpi_notifier((struct notifier_block*)(*nb));
     }
 
     //register notifier chain for ac adapter events
     if (KCL_ACPI_GetDevHandle(device, "_PSR", &dummy_handle) == KCL_ACPI_OK)
     {
-        return register_acpi_notifier(&firegl_acpi_ac_notifier);
+        *nb = kmalloc(sizeof(struct notifier_block), GFP_KERNEL);
+        if (!*nb)
+        {
+            KCL_DEBUG_ERROR ("Could not allocate enough memory for ac notifier_block\n");
+            return -ENOMEM;
+        }
+        ((struct notifier_block*)(*nb))->notifier_call = firegl_acpi_ac_event;
+        return register_acpi_notifier((struct notifier_block*)(*nb));
     }
 
     return 0;
@@ -398,27 +422,33 @@ unsigned int ATI_API_CALL KCL_ACPI_InstallHandler(KCL_ACPI_DevHandle device,
  */
 unsigned int ATI_API_CALL KCL_ACPI_RemoveHandler(KCL_ACPI_DevHandle device,
                                                  unsigned int handler_type,
-                                                 KCL_ACPI_CallbackHandle handler)
+                                                 KCL_ACPI_CallbackHandle handler,
+                                                 KCL_NOTIFIER_BLOCKER *nb)
 {
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
     return acpi_remove_notify_handler(device, handler_type, (acpi_notify_handler)handler);
 #else
     KCL_ACPI_DevHandle dummy_handle;
+    int ret = 0;
 
     //unregister notifier chain for video switching events
     if ((KCL_ACPI_GetDevHandle(device, "_DOD", &dummy_handle) == KCL_ACPI_OK)
         || (KCL_ACPI_GetDevHandle(device, "_DOS", &dummy_handle) == KCL_ACPI_OK))
     {
-        return unregister_acpi_notifier(&firegl_acpi_video_notifier);
+        ret = unregister_acpi_notifier((struct notifier_block*)(*nb));
+        kfree(*nb);
+        *nb = NULL;
     }
 
     //unregister notifier chain for ac adapter events
     if (KCL_ACPI_GetDevHandle(device, "_PSR", &dummy_handle) == KCL_ACPI_OK)
     {
-        return unregister_acpi_notifier(&firegl_acpi_ac_notifier);
+        ret = unregister_acpi_notifier((struct notifier_block*)(*nb));
+        kfree(*nb);
+        *nb = NULL;
     }
 
-    return 0;
+    return ret;
 #endif
 }
 
@@ -639,32 +669,6 @@ void KCL_ACPI_VideoNotify(KCL_ACPI_DevHandle handle, unsigned int event, KCL_ACP
     libip_video_notify(handle, event, data);
 }
 
-/** \brief Compare specified busid with busid of specified acpi handle
- * \param handle acpi handle
- * \param bus_id busid from acpi event
- * \return 0 if equal; 1 otherwise
- * */
-int ATI_API_CALL KCL_ACPI_Cmp_BusID(KCL_ACPI_DevHandle handle, KCL_ACPI_BusId bus_id)
-{
-    acpi_status status;
-    struct acpi_device *tdev;
-    int ret = 1;
-
-    status = acpi_bus_get_device(handle, &tdev);
-    if ( ACPI_SUCCESS(status) ) 
-    {
-        if ( !strcmp(tdev->pnp.bus_id, bus_id) )            
-        {
-            ret = 0;
-        }
-    }
-    else
-    {
-        KCL_DEBUG_ERROR ("Could not find device for handle: %p, status: 0x%x.\n", handle, status);
-    }
-    return (ret);
-}
-
 static unsigned int KCL_ACPI_videoDevice(KCL_ACPI_DevHandle handle)
 {
     KCL_ACPI_DevHandle dummy_handle;
@@ -694,11 +698,11 @@ static unsigned int KCL_ACPI_acDevice(KCL_ACPI_DevHandle handle)
     }
 }
 
-static unsigned int KCL_ACPI_PowerXpressDevice(KCL_ACPI_DevHandle handle)
+unsigned int ATI_API_CALL KCL_ACPI_PowerXpressDevice(KCL_ACPI_DevHandle handle)
 {
-    KCL_ACPI_DevHandle dummy_handle;
+    KCL_ACPI_DevHandle dummy_handle = NULL;
 
-    if (KCL_ACPI_GetDevHandle(handle, "ATPX", &dummy_handle) == KCL_ACPI_OK)
+    if ((KCL_ACPI_GetDevHandle(handle, "ATPX", &dummy_handle) == KCL_ACPI_OK) && dummy_handle)
     {
         return KCL_ACPI_OK;
     }
@@ -776,7 +780,13 @@ static unsigned int KCL_ACPI_SearchHandles(KCL_ACPI_DevHandle handle, unsigned i
 unsigned int ATI_API_CALL KCL_ACPI_GetHandles(kcl_match_info_t *pInfo)
 {
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,12)
-    #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
+    #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
+        #ifdef ACPI_HANDLE
+        pInfo->video_handle = ACPI_HANDLE(&(pInfo->pcidev->dev));
+        #else
+        pInfo->video_handle = pInfo->pcidev->dev.acpi_node.handle;
+        #endif
+    #elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
         pInfo->video_handle = pInfo->pcidev->dev.archdata.acpi_handle;
     #else 
         pInfo->video_handle = pInfo->pcidev->dev.firmware_data;
@@ -819,6 +829,38 @@ KCL_ACPI_DevHandle ATI_API_CALL KCL_ACPI_GetAlternateHandle(KCL_ACPI_DevHandle p
     return retHandle;
 }
 
+static acpi_status KCL_ACPI_Slot_No_Hotplug(KCL_ACPI_DevHandle handle, u32 lvl, void *data, void **rv)
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12,7)
+   struct acpi_device *tdev;
+   struct pci_dev *pdev = (struct pci_dev *)data;
+   int device;
+
+   acpi_bus_get_device(handle, &tdev);
+   if(tdev != NULL)
+   {
+      device = (acpi_device_adr(tdev) >> 16) & 0xffff;
+      if(PCI_SLOT(pdev->devfn) == device)
+      {
+         tdev->flags.no_hotplug = true;
+      }
+   }
+#endif
+   return 0;
+}
+
+void ATI_API_CALL KCL_ACPI_No_Hotplug(void* dev)
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12,7)
+    struct pci_dev  *pdev = (struct pci_dev*)dev;
+
+    if(pdev && pdev->bus && pdev->bus->bridge)
+    {
+       acpi_walk_namespace(ACPI_TYPE_DEVICE, ACPI_HANDLE(pdev->bus->bridge), 1, KCL_ACPI_Slot_No_Hotplug, NULL, pdev , NULL);
+    }
+#endif
+}
+
 #else
 
 /*
@@ -913,7 +955,7 @@ void KCL_ACPI_VideoNotify(KCL_ACPI_DevHandle handle, unsigned int event, KCL_ACP
     return NULL;
 }
 
-int ATI_API_CALL KCL_ACPI_Cmp_BusID(KCL_ACPI_DevHandle handle, KCL_ACPI_BusId bus_id)
+unsigned int ATI_API_CALL KCL_ACPI_PowerXpressDevice(KCL_ACPI_DevHandle handle)
 {
     return (1);
 }
@@ -928,6 +970,11 @@ KCL_ACPI_DevHandle ATI_API_CALL KCL_ACPI_GetAlternateHandle(KCL_ACPI_DevHandle p
     return NULL;
 }
 
+void ATI_API_CALL KCL_ACPI_No_Hotplug(void *)
+{ 
+    return;
+}
+
 #endif
 
 /*
@@ -949,3 +996,34 @@ int ATI_API_CALL KCL_ACPI_GetNewVt(int console)
 #endif
 }
 
+/** \Parse a ACPI table with user defined function
+ *  \param id           ACPI signatures 
+ *  \param handler      Callback function to parse ACPI table
+ *  \return
+ */
+int ATI_API_CALL KCL_ACPI_ParseTable(char *id, KCL_ACPI_IntCallbackHandle handler)
+{
+    struct acpi_table_header *hdr = NULL;
+    acpi_size tbl_size;
+
+    if (!handler)
+    {
+        return KCL_ACPI_ERROR; 
+    }
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,6,3)    
+    if (!ACPI_SUCCESS(acpi_get_table_with_size(id, 0, &hdr, &tbl_size)))
+#else
+    tbl_size = 0x7fffffff;
+    if (!ACPI_SUCCESS(acpi_get_table(id, 0, &hdr)))
+#endif
+    {
+        return KCL_ACPI_ERROR;
+    }
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,1)
+    ((acpi_tbl_table_handler)handler)(hdr);
+#else
+    ((acpi_table_handler)handler)(hdr);
+#endif
+    return KCL_ACPI_OK;
+}
