@@ -37,14 +37,15 @@
 #include <linux/autoconf.h>
 #endif
 #include <linux/pci.h>
+#include <linux/acpi.h>
 
 #include "kcl_config.h"
 #include "kcl_osconfig.h"
 #include "kcl_type.h"
 #include "kcl_iommu.h"
+#include "kcl_acpi.h"
 #include "kcl_debug.h"
 #include "firegl_public.h"
-
 #if defined(CONFIG_AMD_IOMMU_V2) || defined(CONFIG_AMD_IOMMU_V2_MODULE)
 #include <linux/amd-iommu.h>
 #include <linux/kthread.h>
@@ -52,6 +53,9 @@
 
 #endif
 
+//temporarily lay it global in iommu field. keep iommu hack being transparent to driver
+u8 *iommu_base;
+
 /** \brief Registers a peripheral device to the IOMMU driver.  
  *  \param pcidev [in] PCI device handle
  *  \param pPasids [out] num of maximum PASIDs supported.   
@@ -178,4 +182,76 @@ void ATI_API_CALL KCL_IOMMU_UnbindPasid( KCL_PCI_DevHandle pcidev,int pasid)
 #endif    
 }
 
+/** \check whether IOMMU is already initialized
+ *  \ param pcidev [in] PCI device handle
+ */
+int ATI_API_CALL KCL_IOMMU_CheckInfo( KCL_PCI_DevHandle pcidev)
+{
+    struct pci_dev* pdev = (struct pci_dev*)pcidev;
+	
+    // CONFIG_DMAR is before KERNEL_VERSION(3,1,0); CONFIG_INTEL_IOMMU is including and after.
+#if defined(CONFIG_DMAR) || defined(CONFIG_INTEL_IOMMU) || defined(CONFIG_AMD_IOMMU)
+    if ( pdev->dev.archdata.iommu )
+    {
+        return 1;
+    }
+#endif
+    return 0;
+}
+
+static int ATI_API_CALL KCL_IOMMU_MapMMIO(struct acpi_table_header *table)
+{
+    u8 *p = (u8 *)table, *end = (u8 *)table;
+    struct ivhd_header *h;
+    iommu_base = NULL;
+
+    end += table->length;
+    p += IVRS_HEADER_LENGTH;
+
+    while (p < end) {
+        h = (struct ivhd_header *)p;
+    	switch (*p) {
+    	case ACPI_IVHD_TYPE:
+            iommu_base = KCL_IO_MEM_Map(h->mmio_phys, MMIO_REGION_LENGTH, KCL_IOREMAPTYPE_NoCache); 
+		    break;
+	    default:
+		    break;
+	    }
+	    p += h->length;
+    }
+    return 0;
+}
+
+/** \brief set a physical address range not to be translated by IOMMU
+ *  \param pa           physical address 
+ *  \param length       the size of this range 
+ */
+void ATI_API_CALL KCL_IOMMU_SetExclusion(unsigned long long pa, unsigned long long length)
+{
+    u64 start,limit;
+    u64 entry;
+
+    if (KCL_ACPI_ParseTable("IVRS", (KCL_ACPI_IntCallbackHandle)KCL_IOMMU_MapMMIO) != KCL_ACPI_OK)
+    {
+        return;
+    }
+
+    if (!iommu_base)
+    {
+        return;
+    }
+
+    start = pa & PAGE_MASK;
+    limit = (start + length) & PAGE_MASK;
+
+    entry = start | MMIO_EXCL_ENABLE_MASK | MMIO_EXCL_ALLOW_MASK;
+    KCL_IO_MEM_CopyToIO(iommu_base + MMIO_EXCL_BASE_OFFSET, &entry, sizeof(entry));
+
+    entry = limit;
+    KCL_IO_MEM_CopyToIO(iommu_base + MMIO_EXCL_LIMIT_OFFSET, &entry, sizeof(entry));
+
+    KCL_IO_MEM_Unmap(iommu_base);
+}
+
+
 
